/*
 *	input_alac.cpp
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spWave_Caf.h>
#include <sp/spMp4.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include "ALACEncoder.h"
#include "ALACDecoder.h"
#include "ALACBitUtilities.h"

#define SP_ALAC_INPUT_PLUGIN

#include "plugin_alac.cpp"

static const char *sp_alac_file_type_list[] = {
    "ALACMP4",
    "ALACCAF",
    NULL,
};
static const char *sp_alac_file_desc_list[] = {
    "Apple Lossless (MP4)",
    "Apple Lossless (Core Audio Format)",
    NULL,
};
static const char *sp_alac_file_filter_list[] = {
    "*.m4a",
    "*.caf",
    NULL,
};
    
static spBool spIsSupportedByPluginAlac(const char *filename);

static spPluginError spOpenPluginAlac(void *instance, const char *filename, const char *mode);
static spBool spClosePluginAlac(void *instance);
static spPluginState spGetPluginStateAlac(void *instance);
static spLong spGetPluginCurrentPositionAlac(void *instance);
static spBool spStopPluginAlac(void *instance);
static spBool spPausePluginAlac(void *instance);
static spBool spRestartPluginAlac(void *instance);

static long spReadPluginAlac(void *instance, char *data, long length);
static spBool spSeekPluginAlac(void *instance, spLong pos);
static spLong spGetPluginTotalLengthAlac(void *instance);

static spInputPluginRec sp_input_plugin_alac = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "ALAC",
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginAlac,
    spFreePluginAlac,
    "ALAC Input Plugin",
    "ALAC Input Plugin Version 0.1.1\nCopyright(C) Hideki Banno\nBased on the Apple Lossless Audio Codec Project\n(http://alac.macosforge.org/trac/wiki)",
    
    spInitPluginInstanceAlac,
    spFreePluginInstanceAlac,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_alac_file_type_list,
    sp_alac_file_desc_list,
    sp_alac_file_filter_list,
    spIsSupportedByPluginAlac,
    spSetPluginFileTypeAlac,  
    spGetPluginFileTypeAlac,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK
     | SP_SONG_ALBUM_ARTIST_MASK | SP_SONG_COMPOSER_MASK | SP_SONG_TRACK_TOTAL_MASK
     | SP_SONG_DISC_MASK | SP_SONG_DISC_TOTAL_MASK),
    spSetPluginSongInfoAlac,
    spGetPluginSongInfoAlac,
    spGetPluginBestSuffixAlac,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
    
    spSetPluginSampleBitAlac,   
    spGetPluginSampleBitAlac,   
    spSetPluginChannelAlac,     
    spGetPluginChannelAlac,     
    spSetPluginSampleRateAlac,  
    spGetPluginSampleRateAlac,  
    spSetPluginOtherInfoAlac,   
    spGetPluginOtherInfoAlac,
    
    spOpenPluginAlac,              
    spClosePluginAlac,             
    spGetPluginStateAlac,          
    spGetPluginCurrentPositionAlac,
    spStopPluginAlac,              
    spPausePluginAlac,             
    spRestartPluginAlac,

    spReadPluginAlac,             
    spSeekPluginAlac,
    spGetPluginTotalLengthAlac,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_alac;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(input_alac)

static uint8_t *spOpenInputFileAlac(spAlacPluginInstance pinstance, const char *filename, spChunkSize *pMagicCookieSize)
{
    spBool flag;
    spChunkSize nread;
    uint8_t *theMagicCookie = NULL;
    spChunkSize theMagicCookieSize = 0;

    if (pinstance == NULL || pinstance->decoder == NULL) {
	spDebug(10, "spOpenInputFileAlac", "instance has'nt been initialized\n");
	return NULL;
    }
    
    pinstance->fp = NULL;
    pinstance->mp4Trak = NULL;
    pinstance->mp4Stbl = NULL;
    pinstance->mp4Stsc = NULL;
    pinstance->mp4Soun = NULL;
    pinstance->mp4Alac = NULL;
    pinstance->mp4Meta = NULL;
    pinstance->mp4DescriptionIndex = 0;
    pinstance->readBuffer = NULL;
    pinstance->writeBuffer = NULL;
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "rb")) == NULL) {
	spDebug(10, "spOpenInputFileAlac", "open failed\n");
	return NULL;
    }
    spDebug(10, "spOpenInputFileAlac", "filename = %s\n", filename);

    flag = SP_TRUE;
    
    if ((nread = spReadCafHeader(&pinstance->cafHeader, pinstance->fp)) > 0) {
	spCafGeneralDataChunk *kuki;

	pinstance->type_index = SP_ALAC_TYPE_CAF;
        spDebug(10, "spOpenInputFileAlac", "CAF ALAC: %s\n", filename);
	
	if (!strneq(pinstance->cafHeader.desc->mFormatID, "alac", 4)
	    || (kuki = (spCafGeneralDataChunk *)spFindCafChunkFromHeader(&pinstance->cafHeader, "kuki")) == NULL
	    || kuki->mData == NULL) {
	    spDebug(10, "spOpenInputFileAlac", "not ALAC or spFindCafChunkFromHeader failed\n");
	    spFreeCafHeader(&pinstance->cafHeader);
            flag = SP_FALSE;
	} else {

            theMagicCookie = (uint8_t *)kuki->mData;
            theMagicCookieSize = spGetCafChunkContentSize((spCafChunk *)kuki, NULL);

            pinstance->format.mSampleRate = pinstance->cafHeader.desc->mSampleRate;
            memcpy(&pinstance->format.mFormatID, pinstance->cafHeader.desc->mFormatID, 4);
            pinstance->format.mFormatFlags = pinstance->cafHeader.desc->mFormatFlags;
            pinstance->format.mBytesPerPacket = pinstance->cafHeader.desc->mBytesPerPacket;
            pinstance->format.mFramesPerPacket = pinstance->cafHeader.desc->mFramesPerPacket;
            pinstance->format.mChannelsPerFrame = pinstance->cafHeader.desc->mChannelsPerFrame;
            pinstance->format.mBitsPerChannel = pinstance->cafHeader.desc->mBitsPerChannel;
            pinstance->format.mBytesPerFrame = 0;
            pinstance->format.mReserved = 0;
	
            switch (pinstance->cafHeader.desc->mFormatFlags) {
              case SP_CAF_ALAC_FORMAT_FLAG_32BIT_SOURCE_DATA:
                pinstance->bitsPerChannel = 32;
                break;
              case SP_CAF_ALAC_FORMAT_FLAG_24BIT_SOURCE_DATA:
                pinstance->bitsPerChannel = 24;
                break;
              case SP_CAF_ALAC_FORMAT_FLAG_20BIT_SOURCE_DATA:
                pinstance->bitsPerChannel = 20;
                break;
              default:
                pinstance->bitsPerChannel = 16;
                break;
            }

            pinstance->total_duration = spGetCafTotalDuration(&pinstance->cafHeader);
        }
    } else if (
#if 0
        (spEqSuffix(filename, ".m4a") || spEqSuffix(filename, ".m4v")
                || spEqSuffix(filename, ".mp4"))
               &&
#endif
               (nread = spReadMp4Header(&pinstance->mp4Header, pinstance->fp)) > 0) {
	pinstance->type_index = SP_ALAC_TYPE_MP4;
        spDebug(10, "spOpenInputFileAlac", "MP4 ALAC: %s\n", filename);
        
	if ((pinstance->mp4Trak = spFindMp4TrackBox(&pinstance->mp4Header, 0, "soun")) != NULL) {
	    if ((pinstance->mp4Stbl = (spMp4SampleTableBox *)spFindMp4BoxFromHeader(&pinstance->mp4Header, "stbl", "minf")) != NULL) {
		
		pinstance->mp4Stsc = pinstance->mp4Stbl->stsc;
		if ((pinstance->mp4Alac = (spMp4DataOnlyBox *)spFindMp4BoxFromHeader(&pinstance->mp4Header, "alac", "alac")) != NULL) {
		    pinstance->mp4Soun = (spMp4SampleEntryBox *)pinstance->mp4Alac->parent;
		}
	    }
	}

	if (pinstance->mp4Header.moov == NULL
	    || pinstance->mp4Trak == NULL || pinstance->mp4Stbl == NULL || pinstance->mp4Stsc == NULL
	    || pinstance->mp4Alac == NULL || pinstance->mp4Alac->data == NULL || pinstance->mp4Soun == NULL) {
            flag = SP_FALSE;
	    spFreeMp4Header(&pinstance->mp4Header);
	} else {
            theMagicCookie = (uint8_t *)pinstance->mp4Alac->data;
            theMagicCookieSize = spGetMp4BoxContentSize((spMp4Box *)pinstance->mp4Alac, NULL);

            spDebug(10, "spOpenInputFileAlac", "sizeof(spLong32) = %d, samplerate = %f (%lx)\n",
                    sizeof(spLong32), (double)pinstance->mp4Soun->u.soun.samplerate / 65536, pinstance->mp4Soun->u.soun.samplerate);
            pinstance->format.mSampleRate = (double)pinstance->mp4Soun->u.soun.samplerate / 65536.0;
            pinstance->format.mFormatID = 'alac';
            pinstance->format.mFormatFlags = 0;
            pinstance->format.mBytesPerPacket = 0;
            pinstance->format.mFramesPerPacket = spGetMp4MaxSampleDuration(pinstance->mp4Stbl->stts);
            pinstance->format.mChannelsPerFrame = pinstance->mp4Soun->u.soun.channelcount;
            pinstance->format.mBitsPerChannel = 0;
            pinstance->format.mBytesPerFrame = 0;
            pinstance->format.mReserved = 0;
	
            pinstance->bitsPerChannel = pinstance->mp4Soun->u.soun.samplesize;
            pinstance->total_duration = pinstance->mp4Stbl->stts->total_duration;
        }
    } else {
        flag = SP_FALSE;
    }

    if (flag == SP_FALSE) {
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
        return NULL;
    } else {
        pinstance->bytesPerFrame = (pinstance->bitsPerChannel * pinstance->format.mChannelsPerFrame) / 8;
    
        if (pMagicCookieSize != NULL) *pMagicCookieSize = theMagicCookieSize;

        spDebug(10, "spOpenInputFileAlac", "done\n");
    
        return theMagicCookie;
    }
}

static spBool spCloseInputFileAlac(spAlacPluginInstance pinstance)
{
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	
	if (pinstance->readBuffer != NULL) xfree(pinstance->readBuffer);
	pinstance->readBuffer = NULL;
	
	if (pinstance->writeBuffer != NULL) xfree(pinstance->writeBuffer);
	pinstance->writeBuffer = NULL;
	
	if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	    spFreeCafHeader(&pinstance->cafHeader);
	} else {
	    spFreeMp4Header(&pinstance->mp4Header);
	}
	
	pinstance->fp = NULL;
	pinstance->current_pos = 0;
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spIsSupportedByPluginAlac(const char *filename)
{
    spBool flag;
    spAlacPluginInstance pinstance;

    spDebug(10, "spIsSupportedByPluginAlac", "filename = %s\n", filename);
    
    if ((pinstance = (spAlacPluginInstance)spInitPluginInstanceAlac("C")) == NULL) {
	return SP_FALSE;
    }
    
    if (spOpenInputFileAlac(pinstance, filename, NULL) != NULL) {
	spCloseInputFileAlac(pinstance);
	flag = SP_TRUE;
    } else {
	flag = SP_FALSE;
    }
    spFreePluginInstanceAlac(pinstance);
    
    spDebug(10, "spIsSupportedByPluginAlac", "flag = %d\n", flag);
    
    return flag;
}

static spPluginError spOpenPluginAlac(void *instance, const char *filename, const char *mode)
{
    uint8_t *theMagicCookie = NULL;
    spChunkSize theMagicCookieSize = 0;
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    if ((theMagicCookie = spOpenInputFileAlac(pinstance, filename, &theMagicCookieSize)) == NULL) {
	spDebug(10, "spOpenPluginAlac", "spOpenInputFileAlac failed\n");
	spCloseInputFileAlac(pinstance);
	return SP_PLUGIN_ERROR_OPEN;
    }

    if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	spParseCafSongInfo(&pinstance->cafHeader, &pinstance->song_info);
    } else {
	if (pinstance->mp4Header.moov->udta != NULL) {
	    pinstance->mp4Meta = pinstance->mp4Header.moov->udta->meta;
	    spParseMp4SongInfo(pinstance->mp4Meta, &pinstance->song_info);
	}
    }
    
    pinstance->inputPacketBytes = pinstance->format.mChannelsPerFrame
	* (pinstance->bitsPerChannel >> 3) * pinstance->format.mFramesPerPacket + kALACMaxEscapeHeaderBytes;
    pinstance->outputPacketBytes = pinstance->inputPacketBytes - kALACMaxEscapeHeaderBytes;
    pinstance->readBuffer = xalloc(pinstance->inputPacketBytes, uint8_t);
    pinstance->writeBuffer = xalloc(pinstance->outputPacketBytes, uint8_t);
    
    pinstance->decoder->Init(theMagicCookie, theMagicCookieSize);
    
    BitBufferInit(&pinstance->inputBuffer, pinstance->readBuffer, pinstance->inputPacketBytes);

    spSeekPluginAlac(pinstance, 0);

    spDebug(10, "spOpenPluginAlac", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    return spCloseInputFileAlac(pinstance);
}

static spPluginState spGetPluginStateAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    return pinstance->current_pos;
}

static spBool spStopPluginAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginAlac(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginAlac(void *instance)
{

    return SP_TRUE;
}

static spBool spUpdateInputDataPosAlac(spAlacPluginInstance pinstance, spBool *end_flag)
{
    spDebug(80, "spUpdateInputDataPosAlac", "in: pinstance->packetTableCount = %ld\n", pinstance->packetTableCount);

    if (pinstance->numBytes > 0 && pinstance->bufferCopyBytes >= (unsigned long)pinstance->numBytes) {
	pinstance->bufferCopyBytes = 0;
	pinstance->numBytes = 0;
    }
    
    pinstance->prevInputDataPos = pinstance->inputDataPos;
    
    if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	pinstance->inputPacketBytes = spGetCafPacketSize(&pinstance->cafHeader, pinstance->packetTableCount);
	if (pinstance->inputPacketBytes <= 0) {
	    pinstance->numBytes = 0;
	    *end_flag = SP_TRUE;
	    return SP_TRUE;
	}
    } else {
	spFileOffset posDiff;

	pinstance->inputDataPos = spConvertMp4SampleToFileOffset(pinstance->mp4Stsc, pinstance->mp4DescriptionIndex,
								 pinstance->packetTableCount, NULL, NULL,
								 &pinstance->inputPacketBytes);
	if (pinstance->inputPacketBytes <= 0) {
	    pinstance->numBytes = 0;
	    *end_flag = SP_TRUE;
	    return SP_TRUE;
	}

	posDiff = pinstance->inputDataPos - pinstance->prevInputDataPos;
	spDebug(80, "spUpdateInputDataPosAlac", "posDiff = %ld (%ld - %ld)\n",
		(long)posDiff, (long)pinstance->inputDataPos, (long)pinstance->prevInputDataPos);
    
	if (posDiff != 0) {
	    if (spSeekFile(pinstance->fp, posDiff, SEEK_CUR) != 0) {
		return SP_FALSE;
	    }
	}
    }
    pinstance->inputDataPos += pinstance->inputPacketBytes;
    ++pinstance->packetTableCount;

    spDebug(80, "spUpdateInputDataPosAlac", "done: inputDataPos = %ld, inputPacketBytes = %ld\n",
	    (long)pinstance->inputDataPos, pinstance->inputPacketBytes);
    
    return SP_TRUE;
}

static long spDecodeFrameAlac(spAlacPluginInstance pinstance, spBool *end_flag)
{
    long nread;
    int32_t status;

    if (spUpdateInputDataPosAlac(pinstance, end_flag) == SP_FALSE) {
	spDebug(10, "spDecodeFrameAlac", "spUpdateInputDataPosAlac failed\n");
	return -1;
    }
    if (*end_flag == SP_TRUE) {
	return 0;
    }
    
    if ((nread = fread(pinstance->readBuffer, 1, pinstance->inputPacketBytes, pinstance->fp)) != (long)pinstance->inputPacketBytes) {
	spDebug(10, "spDecodeFrameAlac", "fread failed: nread = %ld\n", nread);
	return nread;
    }

    pinstance->numFrames = 0;
    status = pinstance->decoder->Decode(&pinstance->inputBuffer, pinstance->writeBuffer,
					pinstance->format.mFramesPerPacket, pinstance->format.mChannelsPerFrame,
					&pinstance->numFrames);
    pinstance->numBytes = pinstance->numFrames * pinstance->bytesPerFrame;
    
    BitBufferReset(&pinstance->inputBuffer);

    spDebug(80, "spDecodeFrameAlac", "done: status = %ld, numFrames = %ld, numBytes = %ld\n",
	    status, (long)pinstance->numFrames, (long)pinstance->numBytes);
    
    return nread;
}

static long spReadPluginAlac(void *instance, char *data, long length)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    unsigned long bytes;
    unsigned long ncopy;
    unsigned long nremain;
    unsigned long nframe;
    spBool end_flag;

    if (pinstance->fp == NULL) return -1;

    if (pinstance->current_pos >= pinstance->total_duration) {
	return 0;
    }

    spDebug(80, "spReadPluginAlac", "length = %ld, current_pos = %ld\n",
	    length, (long)pinstance->current_pos);
    
    nremain = pinstance->bytesPerFrame * (length / pinstance->format.mChannelsPerFrame);
    ncopy = 0;
    end_flag = SP_FALSE;

    while (nremain > 0) {
	if (pinstance->bufferCopyBytes >= (unsigned long)pinstance->numBytes) {
	    if (spDecodeFrameAlac(pinstance, &end_flag) < 0) {
		return -1;
	    }
	    if (end_flag == SP_TRUE) {
		break;
	    }
	}

	spDebug(100, "spReadPluginAlac", "numBytes = %ld, bufferCopyBytes = %ld\n",
		(long)pinstance->numBytes, pinstance->bufferCopyBytes);
	if ((unsigned long)pinstance->numBytes > pinstance->bufferCopyBytes) {
	    bytes = MIN(nremain, (unsigned long)pinstance->numBytes - pinstance->bufferCopyBytes);
	    memcpy(data + ncopy, pinstance->writeBuffer + pinstance->bufferCopyBytes, bytes);
	    ncopy += bytes;
	    nremain -= bytes;
	
	    pinstance->bufferCopyBytes += bytes;
	}
    }

    nframe = ncopy / pinstance->bytesPerFrame;
    pinstance->current_pos += nframe;
    spDebug(80, "spReadPluginAlac", "done: nframe = %ld, current_pos = %ld / %ld\n",
	    nframe, (long)pinstance->current_pos, (long)pinstance->total_duration);

    return nframe * pinstance->format.mChannelsPerFrame;
}

static spBool spSeekPluginAlac(void *instance, spLong pos)
{
    spFileOffset offset;
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    if (pinstance->fp == NULL) return SP_FALSE;

    spDebug(80, "spSeekPluginAlac", "pos = %ld\n", (long)pos);
    
    pinstance->numBytes = 0;
    pinstance->prevInputDataPos = 0;
    
    if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	long packet, remainder = 0;
	
	offset = spConvertCafTimeToFileOffset(&pinstance->cafHeader, pos,
					      &packet, &remainder, NULL);
	
	pinstance->packetTableCount = packet;
	pinstance->bufferCopyBytes = remainder * pinstance->bytesPerFrame;
	pinstance->inputDataPos = offset;
	spDebug(10, "spSeekPluginAlac", "packet = %ld, remainder = %ld\n", packet, remainder);
    } else {
	unsigned long remainder = 0;

	offset = 0;
	pinstance->packetTableCount = spConvertMp4TimeToSample(pinstance->mp4Stbl->stts, pos, &remainder);
	pinstance->bufferCopyBytes = remainder * pinstance->bytesPerFrame;
	pinstance->inputDataPos = 0;
    }
    
    spDebug(80, "spSeekPluginAlac", "offset = %ld, packetTableCount = %ld, bufferCopyBytes = %ld\n",
	    (long)offset, pinstance->packetTableCount, pinstance->bufferCopyBytes);
    
    if (spSeekFile(pinstance->fp, offset, SEEK_SET) != 0) {
	spDebug(10, "spSeekPluginAlac", "spSeekFile failed\n");
	return SP_FALSE;
    }
    
    pinstance->current_pos = pos;
    
    return SP_TRUE;
}

static spLong spGetPluginTotalLengthAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    return pinstance->total_duration;
}

