/*
 *	input_nist.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include "plugin_nist.c"

static const char *sp_nist_file_type_list[] = {
    "nist-little",
    "nist-big",
    "nist-ulaw",
    "nist-alaw",
    NULL,
};
static const char *sp_nist_file_desc_list[] = {
    "NIST SPHERE (Little Endian)",
    "NIST SPHERE (Big Endian)",
    "NIST SPHERE (U-law)",
    "NIST SPHERE (A-law)",
    NULL,
};
static const char *sp_nist_file_filter_list[] = {
    "*.wav",
    "*.wav",
    "*.wav",
    "*.wav",
    NULL,
};
    
static spBool spIsSupportedByPluginNist(const char *filename);

static spPluginError spOpenPluginNist(void *instance, const char *filename, const char *mode);
static spBool spClosePluginNist(void *instance);
static spPluginState spGetPluginStateNist(void *instance);
static spLong spGetPluginCurrentPositionNist(void *instance);
static spBool spStopPluginNist(void *instance);
static spBool spPausePluginNist(void *instance);
static spBool spRestartPluginNist(void *instance);

static long spReadPluginNist(void *instance, char *data, long length);
static spBool spSeekPluginNist(void *instance, spLong pos);
static spLong spGetPluginTotalLengthNist(void *instance);

static spInputPluginRec sp_input_plugin_nist = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "NIST",
    2,
    SP_PLUGIN_PRIORITY_HIGHEST,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginNist,
    spFreePluginNist,
    "NIST Input Plugin",
    "NIST Input Plugin Version 0.2\nCopyright(C) Hideki Banno",
    
    spInitPluginInstanceNist,
    spFreePluginInstanceNist,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_nist_file_type_list,
    sp_nist_file_desc_list,
    sp_nist_file_filter_list,
    spIsSupportedByPluginNist,
    spSetPluginFileTypeNist,  
    spGetPluginFileTypeNist,
    0,
    spSetPluginSongInfoNist,
    spGetPluginSongInfoNist,
    spGetPluginBestSuffixNist,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
    
    spSetPluginSampleBitNist,   
    spGetPluginSampleBitNist,   
    spSetPluginChannelNist,     
    spGetPluginChannelNist,     
    spSetPluginSampleRateNist,  
    spGetPluginSampleRateNist,  
    spSetPluginOtherInfoNist,   
    spGetPluginOtherInfoNist,
    
    spOpenPluginNist,              
    spClosePluginNist,             
    spGetPluginStateNist,          
    spGetPluginCurrentPositionNist,
    spStopPluginNist,              
    spPausePluginNist,             
    spRestartPluginNist,

    spReadPluginNist,             
    spSeekPluginNist,
    spGetPluginTotalLengthNist,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_nist;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(input_nist)

static int decodeNistHeader(spNistPluginInstance pinstance, FILE *fp)
{
    int head_len = -1;
    int buf_size = 1024;
    char buf[1024];
    char id[1024];
    char opt[1024];
    char val[1024];

    spSeekFile(fp, 0, SEEK_SET);
    
    if (fgetsn(buf, buf_size, fp) == NULL || !strneq(buf, "NIST_1A", 7)) {
	return -1;
    }
    if (fgetsn(buf, buf_size, fp) == NULL || sscanf(buf, "%d", &head_len) < 1) {
	return -1;
    }
    spDebug(30, "decodeNistHeader", "head_len = %d\n", head_len);
    
    while (fgetsn(buf, buf_size, fp) != NULL) {
	strcpy(id, ""); strcpy(opt, ""); strcpy(val, "");
	if (sscanf(buf, "%s %s %s", id, opt, val) >= 1) {
	    if (streq(id, "end_head")) {
		break;
	    } else if (streq(id, "channel_count")) {
		pinstance->num_channel = atoi(val);
	    } else if (streq(id, "sample_rate")) {
		pinstance->samp_rate = atof(val);
	    } else if (streq(id, "sample_n_bytes")) {
		pinstance->samp_bit = atoi(val) * 8;
	    } else if (streq(id, "sample_sig_bits")) {
		pinstance->samp_bit = atoi(val);
	    } else if (streq(id, "sample_coding")) {
		if (streq(val, "ulaw")) {
		    strcpy(pinstance->type, "nist-ulaw");
		} else if (streq(val, "alaw")) {
		    strcpy(pinstance->type, "nist-alaw");
		} else if (!streq(val, "pcm")) {
		    return -1;
		}
	    } else if (streq(id, "sample_byte_format")) {
		if (!streq(pinstance->type, "nist-ulaw") && !streq(pinstance->type, "nist-alaw")) {
		    if (streq(val, "01")) {
			strcpy(pinstance->type, "nist-little");
		    } else if (streq(val, "10")) {
			strcpy(pinstance->type, "nist-big");
		    } else {
			strcpy(pinstance->type, "unknown");
		    }
		}
	    }
	}
    }
    spDebug(20, "decodeNistHeader", "samp_bit = %d\n", pinstance->samp_bit);
    
    if (pinstance->samp_bit != 8 && pinstance->samp_bit != 16) {
	return -1;
    }
    if (streq(pinstance->type, "unknown")) {
	if (pinstance->samp_bit == 8) {
	    strcpy(pinstance->type, "nist-little");
	} else {
	    return -1;
	}
    }

    if ((pinstance->type_index =
	 spFindPluginRecFileTypeIndex((spIoPluginRec *)&sp_input_plugin_nist, pinstance->type)) < 0) {
	return -1;
    }
    spDebug(20, "decodeNistHeader", "type_index = %d\n", pinstance->type_index);

    return head_len;
}

static spBool spIsSupportedByPluginNist(const char *filename)
{
    spBool flag = SP_FALSE;
    spNistPluginInstance instance;
    FILE *fp;

    instance = (spNistPluginInstance)spInitPluginInstanceNist("C");
    
    if ((fp = spOpenFile(filename, "rb")) != NULL) {
	if (decodeNistHeader(instance, fp) >= 0) {
	    flag = SP_TRUE;
	}
	spCloseFile(fp);
    }
    
    spFreePluginInstanceNist(instance);

    return flag;
}

static spPluginError spOpenPluginNist(void *instance, const char *filename, const char *mode)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "rb")) == NULL) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(10, "spOpenPluginNist", "filename = %s\n", filename);

    if ((pinstance->head_len = decodeNistHeader(pinstance, pinstance->fp)) < 0) {
	spCloseFile(pinstance->fp); pinstance->fp = NULL;
	return SP_PLUGIN_ERROR_FAILURE;
    }
    
    /* get file length */
    pinstance->file_length = getfilesize(filename, pinstance->head_len)
	/ (spLong)pinstance->num_channel / (spLong)(pinstance->samp_bit / 8);
    
    spDebug(10, "spOpenPluginNist", "head_len = %d\n", pinstance->head_len);
    
    /* skip header */
    spSeekFile(pinstance->fp, pinstance->head_len, SEEK_SET);
    pinstance->current_pos = 0;

    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginNist(void *instance)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
	pinstance->head_len = 0;
	pinstance->current_pos = 0;
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateNist(void *instance)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionNist(void *instance)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    return pinstance->current_pos;
}

static spBool spStopPluginNist(void *instance)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	spSeekFile(pinstance->fp, pinstance->head_len, SEEK_SET);
    }
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginNist(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginNist(void *instance)
{

    return SP_TRUE;
}

static long spReadPluginNist(void *instance, char *data, long length)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    long nread;
    spBool swap_flag = SP_FALSE;

    if (pinstance->fp == NULL) return -1;

#if BYTE_ORDER == BIG_ENDIAN
    if (pinstance->type_index == SP_NIST_TYPE_LITTLE) {
	swap_flag = SP_TRUE;
    }
#else
    if (pinstance->type_index == SP_NIST_TYPE_BIG) {
	swap_flag = SP_TRUE;
    }
#endif
    
    if (pinstance->samp_bit == 8) {
	if (pinstance->type_index == SP_NIST_TYPE_ULAW) {
	    nread = freadulaw((short *)data, length, pinstance->fp);
	} else if (pinstance->type_index == SP_NIST_TYPE_ALAW) {
	    nread = freadalaw((short *)data, length, pinstance->fp);
	} else {
	    nread = freadbyte((short *)data, length, pinstance->fp);
	}
    } else if (pinstance->samp_bit == 64) {
	nread = freaddouble((double *)data, length, swap_flag, pinstance->fp);
    } else {
	nread = freadshort((short *)data, length, swap_flag, pinstance->fp);
    }
    pinstance->current_pos += (nread / pinstance->num_channel);
    
    return nread;
}

static spBool spSeekPluginNist(void *instance, spLong pos)
{
    long k, nloop;
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;
    
    if (pinstance->fp == NULL) return SP_FALSE;
    
    spSeekFile(pinstance->fp, pinstance->head_len, SEEK_SET);
    
    /* for a long file */
    nloop = (long)pinstance->num_channel * (long)(pinstance->samp_bit / 8);
    for (k = 0; k < nloop; k++) {
	spSeekFile(pinstance->fp, pos, SEEK_CUR);
    }
    pinstance->current_pos = pos;
    
    return SP_TRUE;
}

static spLong spGetPluginTotalLengthNist(void *instance)
{
    spNistPluginInstance pinstance = (spNistPluginInstance)instance;

    return pinstance->file_length;
}

