/*
 *	output_alac.cpp
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spWave_Caf.h>
#include <sp/spMp4.h>

#include <sp/spOutputPluginP.h>
#include <sp/spPluginMain.h>

#include "ALACEncoder.h"
#include "ALACDecoder.h"
#include "ALACBitUtilities.h"

#define SP_ALAC_OUTPUT_PLUGIN

#include "plugin_alac.cpp"

#define SP_ALAC_MP4_MOOV_SIZE 20432
#define SP_ALAC_MP4_USER_DATA_META_SIZE 4096

#define SP_ALAC_CAF_INFORMATION_CHUNK_SIZE 2048

static const char *sp_alac_file_type_list[] = {
    "ALACMP4",
    "ALACCAF",
    NULL,
};
static const char *sp_alac_file_desc_list[] = {
    "Apple Lossless (MP4)",
    "Apple Lossless (Core Audio Format)",
    NULL,
};
static const char *sp_alac_file_filter_list[] = {
    "*.m4a",
    "*.caf",
    NULL,
};
    
static spBool spIsSupportedByPluginAlac(const char *filename);

static spPluginError spOpenPluginAlac(void *instance, const char *filename, const char *mode);
static spBool spClosePluginAlac(void *instance);
static spPluginState spGetPluginStateAlac(void *instance);
static spLong spGetPluginCurrentPositionAlac(void *instance);
static spBool spStopPluginAlac(void *instance);
static spBool spPausePluginAlac(void *instance);
static spBool spRestartPluginAlac(void *instance);

static long spWritePluginAlac(void *instance, char *data, long length);
static spBool spFlushPluginAlac(void *instance);
static spBool spSetPluginTotalLengthAlac(void *instance, spLong total_length);

static spOutputPluginRec sp_output_plugin_alac = {
    NULL,
    NULL,

    SP_PLUGIN_OUTPUT,
    "ALAC",
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginAlac,
    spFreePluginAlac,
    "ALAC Output Plugin",
    "ALAC Output Plugin Version 0.1\nCopyright(C) Hideki Banno\nBased on the Apple Lossless Audio Codec Project\n(http://alac.macosforge.org/trac/wiki)",
    
    spInitPluginInstanceAlac,
    spFreePluginInstanceAlac,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_alac_file_type_list,
    sp_alac_file_desc_list,
    sp_alac_file_filter_list,
    spIsSupportedByPluginAlac,
    spSetPluginFileTypeAlac,  
    spGetPluginFileTypeAlac,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK
     | SP_SONG_ALBUM_ARTIST_MASK | SP_SONG_COMPOSER_MASK | SP_SONG_TRACK_TOTAL_MASK
     | SP_SONG_DISC_MASK | SP_SONG_DISC_TOTAL_MASK),
    spSetPluginSongInfoAlac,
    spGetPluginSongInfoAlac,
    spGetPluginBestSuffixAlac,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
    
    spSetPluginSampleBitAlac,   
    spGetPluginSampleBitAlac,   
    spSetPluginChannelAlac,     
    spGetPluginChannelAlac,     
    spSetPluginSampleRateAlac,  
    spGetPluginSampleRateAlac,  
    spSetPluginOtherInfoAlac,   
    spGetPluginOtherInfoAlac,
    
    spOpenPluginAlac,              
    spClosePluginAlac,             
    spGetPluginStateAlac,          
    spGetPluginCurrentPositionAlac,
    spStopPluginAlac,              
    spPausePluginAlac,             
    spRestartPluginAlac,

    spWritePluginAlac,             
    spFlushPluginAlac,
    spSetPluginTotalLengthAlac,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_output_plugin_alac;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(output_alac)

typedef uint32_t CAFFChannelLayoutTag;
enum
{
    kCAFFChannelLayoutTag_Mono          = (100<<16) | 1,    // C
    kCAFFChannelLayoutTag_Stereo        = (101<<16) | 2,	// L R
    kCAFFChannelLayoutTag_MPEG_3_0_B    = (113<<16) | 3,	// C L R
    kCAFFChannelLayoutTag_MPEG_4_0_B    = (116<<16) | 4,	// C L R Cs
    kCAFFChannelLayoutTag_MPEG_5_0_D    = (120<<16) | 5,    // C L R Ls Rs
    kCAFFChannelLayoutTag_MPEG_5_1_D    = (124<<16) | 6,	// C L R Ls Rs LFE
    kCAFFChannelLayoutTag_AAC_6_1       = (142<<16) | 7,	// C L R Ls Rs Cs LFE
    kCAFFChannelLayoutTag_MPEG_7_1_B	= (127<<16) | 8     // C Lc Rc L R Ls Rs LFE
};

static CAFFChannelLayoutTag spALACCAFFChannelLayoutTags[kALACMaxChannels] =
{
    kCAFFChannelLayoutTag_Mono,         // C
    kCAFFChannelLayoutTag_Stereo,	// L R
    kCAFFChannelLayoutTag_MPEG_3_0_B,	// C L R
    kCAFFChannelLayoutTag_MPEG_4_0_B,	// C L R Cs
    kCAFFChannelLayoutTag_MPEG_5_0_D,	// C L R Ls Rs
    kCAFFChannelLayoutTag_MPEG_5_1_D,	// C L R Ls Rs LFE
    kCAFFChannelLayoutTag_AAC_6_1,	// C L R Ls Rs Cs LFE
    kCAFFChannelLayoutTag_MPEG_7_1_B	// C Lc Rc L R Ls Rs LFE    (doc: IS-13818-7 MPEG2-AAC)
};

static ALACChannelLayoutTag spGetALACChannelLayoutTag(uint32_t channelsPerFrame)
{
    return spALACCAFFChannelLayoutTags[channelsPerFrame - 1];
}

static spMp4BoxSize spMp4BoxUpdateFuncAlac(spMp4Box *box, void *data)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)data;
    
    spDebug(10, "spMp4BoxUpdateFuncAlac", "type = %c%c%c%c\n",
	    box->header.type[0], box->header.type[1], box->header.type[2], box->header.type[3]);
    

    return 0;
}

static long spOpenOutputFileAlac(spAlacPluginInstance pinstance, const char *filename)
{
    long nwrite = 0;
    uint8_t *theMagicCookie = NULL;
    uint32_t theMagicCookieSize = 0;

    if (pinstance == NULL || pinstance->encoder == NULL) {
	spDebug(10, "spOpenOutputFileAlac", "instance has'nt been initialized\n");
	return 0;
    }
    
    pinstance->totalDataBytes = 0;
    pinstance->fp = NULL;
    pinstance->mp4Trak = NULL;
    pinstance->mp4Stbl = NULL;
    pinstance->mp4Stsc = NULL;
    pinstance->mp4Soun = NULL;
    pinstance->mp4Alac = NULL;
    pinstance->mp4Meta = NULL;
    pinstance->mp4DescriptionIndex = 1;
    pinstance->mp4ChunkOffset = 0;
    pinstance->mp4SamplesPerChunk = 0;
    pinstance->mp4PausedBox = NULL;
    pinstance->readBuffer = NULL;
    pinstance->writeBuffer = NULL;
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "w+b")) == NULL) {
	spDebug(10, "spOpenOutputFileAlac", "open failed\n");
	return 0;
    }

    if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	if (spEqSuffix(filename, ".m4a") == SP_TRUE || spEqSuffix(filename, ".mp4") == SP_TRUE
	    || spEqSuffix(filename, ".m4v") == SP_TRUE || spEqSuffix(filename, ".alac") == SP_TRUE) {
	    pinstance->type_index = SP_ALAC_TYPE_MP4;
	}
    } else {
	if (spEqSuffix(filename, ".caf") == SP_TRUE) {
	    pinstance->type_index = SP_ALAC_TYPE_CAF;
	}
    }
    spDebug(10, "spOpenOutputFileAlac", "filename = %s, type_index = %d\n", filename, pinstance->type_index);
    
    memcpy(&pinstance->inputFormat, &pinstance->format, sizeof(AudioFormatDescription));
    pinstance->inputFormat.mFormatID = kALACFormatLinearPCM;
    if (pinstance->bitsPerChannel == 33) {
	pinstance->inputFormat.mFormatFlags = kALACFormatFlagIsFloat | kALACFormatFlagsNativeEndian;
	pinstance->bitsPerChannel = 32;
    } else {
	pinstance->inputFormat.mFormatFlags = kALACFormatFlagIsSignedInteger /*| kALACFormatFlagIsPacked*/ | kALACFormatFlagsNativeEndian;
    }
    pinstance->inputFormat.mBitsPerChannel = pinstance->bitsPerChannel;
    pinstance->inputFormat.mBytesPerPacket = pinstance->inputFormat.mBytesPerFrame
	= (pinstance->inputFormat.mBitsPerChannel * pinstance->inputFormat.mChannelsPerFrame) / 8;
    pinstance->inputFormat.mFramesPerPacket = 1;
    pinstance->inputFormat.mReserved = 0;
    
    pinstance->format.mFormatID = kALACFormatAppleLossless;
    pinstance->format.mBitsPerChannel = 0;
    pinstance->format.mBytesPerPacket = 0;
    pinstance->format.mBytesPerFrame = 0;
    pinstance->format.mFramesPerPacket = kALACDefaultFramesPerPacket;
    pinstance->format.mReserved = 0;
    spDebug(10, "spOpenOutputFileAlac", "mSampleRate = %f, mChannelsPerFrame = %ld, mFormatFlags = %ld\n",
	    pinstance->format.mSampleRate, pinstance->format.mChannelsPerFrame, pinstance->format.mFormatFlags);

    pinstance->encoder->SetFrameSize(pinstance->format.mFramesPerPacket);
    pinstance->encoder->InitializeEncoder(pinstance->format);
    
    theMagicCookieSize = pinstance->encoder->GetMagicCookieSize(pinstance->format.mChannelsPerFrame);
    theMagicCookie = (uint8_t *)calloc(theMagicCookieSize, 1);
    pinstance->encoder->GetMagicCookie(theMagicCookie, &theMagicCookieSize);
    
    if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	if (nwrite >= 0) {
	    long maxPacketSize;

	    spInitCafHeader(&pinstance->cafHeader, SP_CAF_SUPPORTED_FILE_VERSION, SP_CAF_SUPPORTED_FILE_FLAGS);
	    spUpdateCafAudioDescChunk(&pinstance->cafHeader,
				      pinstance->format.mSampleRate,
				      "alac", pinstance->format.mFormatFlags,
				      pinstance->format.mBytesPerPacket,
				      pinstance->format.mFramesPerPacket,
				      pinstance->format.mChannelsPerFrame,
				      pinstance->bitsPerChannel);
	    spUpdateCafGeneralDataChunk(&pinstance->cafHeader, "kuki", (char *)theMagicCookie, theMagicCookieSize, SP_TRUE);
	    spUpdateCafPacketTableChunkForFixedFrame(&pinstance->cafHeader, kALACMaxEscapeHeaderBytes, pinstance->total_duration, &maxPacketSize);

	    spCreateCafInformationChunk(&pinstance->cafHeader, SP_ALAC_CAF_INFORMATION_CHUNK_SIZE);
	    spAppendCafSongInfo(&pinstance->cafHeader, &pinstance->song_info);

	    spUpdateCafAudioDataChunk(&pinstance->cafHeader, 1, 0);

	    nwrite = (long)spWriteCafHeader(&pinstance->cafHeader, pinstance->fp);

	    if (nwrite <= 0) {
		spDebug(10, "spOpenOutputFileAlac", "spWriteCafHeader failed: nwrite = %ld\n", nwrite);
		spFreeCafHeader(&pinstance->cafHeader);
	    }
	}
    } else {
	spInitMp4Header(&pinstance->mp4Header, "M4A ", 0, 4, "M4A mp42isom");
	pinstance->mp4Trak = spAppendMp4SoundTrackBox(&pinstance->mp4Header, 1, "alac", 1,
						      (unsigned short)pinstance->format.mChannelsPerFrame,
						      (unsigned short)pinstance->bitsPerChannel,
						      0, pinstance->format.mSampleRate, 0, SP_FALSE,
						      spMp4BoxUpdateFuncAlac, pinstance);
	spAppendMp4DataReferenceEntry(pinstance->mp4Trak->mdia->minf->dinf->dref, "url ", "", 0);
	pinstance->mp4Soun = spGetMp4TrackSampleEntryBox(pinstance->mp4Trak, 0);
	pinstance->mp4Alac = (spMp4DataOnlyBox *)spAppendMp4DataOnlyBox((spMp4Box *)pinstance->mp4Soun, "alac", (char *)theMagicCookie, theMagicCookieSize);
	
	pinstance->mp4Stbl = pinstance->mp4Trak->mdia->minf->stbl;
	pinstance->mp4Stsc = pinstance->mp4Stbl->stsc;

	pinstance->mp4Meta = spAppendMp4UserDataMetaBox((spMp4Box *)pinstance->mp4Header.moov,
							SP_ALAC_MP4_USER_DATA_META_SIZE, "mdir", NULL);
	spAppendMp4SongInfo(pinstance->mp4Meta, &pinstance->song_info);

	nwrite = (long)spWriteMp4Header(&pinstance->mp4Header, SP_ALAC_MP4_MOOV_SIZE, &pinstance->mp4PausedBox, pinstance->fp);
	pinstance->mp4ChunkOffset = nwrite;
	
	/*spShowMp4HeaderInfo(&pinstance->mp4Header);*/
    }

    xfree(theMagicCookie);

    if (nwrite > 0) {
	pinstance->bytesPerFrame = (pinstance->format.mChannelsPerFrame * pinstance->bitsPerChannel) / 8;
    }
    
    spDebug(10, "spOpenOutputFileAlac", "done: nwrite = %ld\n", nwrite);
    
    return nwrite;
}

static spBool spCloseOutputFileAlac(spAlacPluginInstance pinstance)
{
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	
	if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	    spFreeCafHeader(&pinstance->cafHeader);
	} else {
	    spFreeMp4Header(&pinstance->mp4Header);
	}
	
	pinstance->fp = NULL;
	pinstance->current_pos = 0;
	pinstance->mp4ChunkOffset = 0;
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static long spEncodeFrameAlac(spAlacPluginInstance pinstance, long *pRemainBytes, spBool last_frame)
{
    long nwrite;
    long remainBytes;

    remainBytes = pinstance->inputPacketBytes - pinstance->bufferCopyBytes;
    if (pRemainBytes != NULL) *pRemainBytes = remainBytes;

    pinstance->numBytes = pinstance->bufferCopyBytes;
    spDebug(10, "spEncodeFrameAlac", "encode numBytes = %ld\n", (long)pinstance->numBytes);
    pinstance->encoder->Encode(pinstance->inputFormat, pinstance->inputFormat, pinstance->readBuffer, pinstance->writeBuffer,
			       &pinstance->numBytes);
    
    if ((nwrite = fwrite(pinstance->writeBuffer, 1, pinstance->numBytes, pinstance->fp)) != (long)pinstance->numBytes) {
	spDebug(10, "spEncodeFrameAlac", "fwrite failed: nwrite = %ld\n", nwrite);
	return nwrite;
    }
    pinstance->totalDataBytes += nwrite;
    
    if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	spSetCafPacketSize(&pinstance->cafHeader, pinstance->packetTableCount, pinstance->numBytes);
    } else {
	spAppendMp4SampleSize(pinstance->mp4Stbl->stsz, pinstance->numBytes);
	spAppendMp4SampleDuration(pinstance->mp4Stbl->stts, pinstance->packetTableCount,
				  pinstance->bufferCopyBytes / pinstance->bytesPerFrame);

	pinstance->mp4SamplesPerChunk++;
	spDebug(80, "spEncodeFrameAlac", "packetTableCount = %ld, mp4SamplesPerChunk = %ld, pinstance->mp4ChunkOffset = %ld\n",
		pinstance->packetTableCount, pinstance->mp4SamplesPerChunk, (long)pinstance->mp4ChunkOffset);
	
	if (last_frame == SP_TRUE || pinstance->mp4SamplesPerChunk == 5) {
	    unsigned long chunk;
	    
	    chunk = 1 + (pinstance->packetTableCount / 5); /* 1-based chunk index */
	    spAppendMp4SampleToChunk(pinstance->mp4Stsc, chunk, pinstance->mp4SamplesPerChunk, pinstance->mp4DescriptionIndex);
	    pinstance->mp4SamplesPerChunk = 0;
	} else if (pinstance->mp4SamplesPerChunk == 1) {
	    spAppendMp4ChunkOffset(pinstance->mp4Stbl->stco, pinstance->mp4ChunkOffset);
	}

	pinstance->mp4ChunkOffset += nwrite;
    }
    
    pinstance->bufferCopyBytes = 0;
    ++pinstance->packetTableCount;

    spDebug(80, "spEncodeFrameAlac", "encode done: numBytes = %ld, nwrite = %ld\n", (long)pinstance->numBytes, nwrite);
    
    return nwrite;
}

static spBool spIsSupportedByPluginAlac(const char *filename)
{
    const char *p;
    spBool flag = SP_FALSE;
    
    if (filename != NULL && (p = strrchr(filename, '.')) != NULL) {
	if (strcaseeq(p, ".caf") || strcaseeq(p, ".m4a") || strcaseeq(p, ".mp4") || strcaseeq(p, ".m4v")
	    || strcaseeq(p, ".alac")) {
	    flag = SP_TRUE;
	}
    }
    
    return flag;
}

static spPluginError spOpenPluginAlac(void *instance, const char *filename, const char *mode)
{
    long nwrite;
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    if ((nwrite = spOpenOutputFileAlac(pinstance, filename)) <= 0) {
	spCloseOutputFileAlac(pinstance);
	spDebug(10, "spOpenPluginAlac", "spOpenOutputFileAlac failed: nwrite = %ld\n", nwrite);
	return SP_PLUGIN_ERROR_OPEN;
    }

    pinstance->inputPacketBytes = pinstance->bytesPerFrame * pinstance->format.mFramesPerPacket;
    pinstance->outputPacketBytes = pinstance->inputPacketBytes + kALACMaxEscapeHeaderBytes;
    pinstance->readBuffer = xalloc(pinstance->inputPacketBytes, uint8_t);
    pinstance->writeBuffer = xalloc(pinstance->outputPacketBytes, uint8_t);
    pinstance->bufferCopyBytes = 0;
    
    pinstance->packetTableCount = 0;
    
    spDebug(10, "spOpenPluginAlac", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginAlac(void *instance)
{
    long nwrite = 0;
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	spBool insufficientBuffer;
	
	if (pinstance->bufferCopyBytes > 0) {
	    long remainBytes;
	    long frames;
	    
	    spEncodeFrameAlac(pinstance, &remainBytes, SP_TRUE);
	    frames = (pinstance->inputPacketBytes - remainBytes) / pinstance->bytesPerFrame;
	}
	
	pinstance->total_duration = pinstance->current_pos;
	
	if (pinstance->type_index == SP_ALAC_TYPE_CAF) {
	    spUpdateCafPacketTableChunkNumberPackets(&pinstance->cafHeader, pinstance->packetTableCount,
						     pinstance->total_duration, &insufficientBuffer);
	    nwrite = (long)spRewriteCafHeader(&pinstance->cafHeader, 1, pinstance->totalDataBytes, pinstance->fp);
	} else {
	    nwrite = (long)spRewriteMp4Header(&pinstance->mp4Header, pinstance->totalDataBytes, &pinstance->mp4PausedBox, pinstance->fp);
	    /*spShowMp4HeaderInfo(&pinstance->mp4Header);*/
	}
	spDebug(50, "spClosePluginAlac", "nwrite = %ld\n", nwrite);

	if (pinstance->readBuffer != NULL) xfree(pinstance->readBuffer);
	pinstance->readBuffer = NULL;
	
	if (pinstance->writeBuffer != NULL) xfree(pinstance->writeBuffer);
	pinstance->writeBuffer = NULL;

	spCloseOutputFileAlac(pinstance);
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    return (spLong)pinstance->current_pos;
}

static spBool spStopPluginAlac(void *instance)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginAlac(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginAlac(void *instance)
{

    return SP_TRUE;
}

static long spWritePluginAlac(void *instance, char *data, long length)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;
    unsigned long bytes;
    unsigned long ncopy;
    unsigned long nremain;
    unsigned long nframe;

    if (pinstance->fp == NULL) return -1;

    if (pinstance->current_pos >= pinstance->total_duration) {
	return 0;
    }

    spDebug(80, "spWritePluginAlac", "length = %ld, current_pos = %ld\n",
	    length, (long)pinstance->current_pos);
    
    nremain = pinstance->bytesPerFrame * (length / pinstance->format.mChannelsPerFrame);
    ncopy = 0;

    while (nremain > 0) {
	bytes = MIN(nremain, pinstance->inputPacketBytes - pinstance->bufferCopyBytes);
	memcpy(pinstance->readBuffer + pinstance->bufferCopyBytes, data + ncopy, bytes);
	ncopy += bytes;
	nremain -= bytes;
	
	pinstance->bufferCopyBytes += bytes;
	
	spDebug(100, "spWritePluginAlac", "ncopy = %ld, numBytes = %ld, bufferCopyBytes = %ld\n",
		ncopy, (long)pinstance->numBytes, pinstance->bufferCopyBytes);
	
	if (pinstance->bufferCopyBytes >= pinstance->inputPacketBytes) {
	    if (spEncodeFrameAlac(pinstance, NULL, SP_FALSE) < 0) {
		return -1;
	    }
	}
    }

    nframe = ncopy / pinstance->bytesPerFrame;
    pinstance->current_pos += nframe;
    spDebug(80, "spWritePluginAlac", "done: nframe = %ld, current_pos = %ld / %ld\n",
	    nframe, (long)pinstance->current_pos, (long)pinstance->total_duration);

    return nframe * pinstance->format.mChannelsPerFrame;
}

static spBool spFlushPluginAlac(void *instance)
{
    return SP_TRUE;
}

static spBool spSetPluginTotalLengthAlac(void *instance, spLong total_length)
{
    spAlacPluginInstance pinstance = (spAlacPluginInstance)instance;

    pinstance->total_duration = total_length;
    
    return SP_TRUE;
}

